import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestLinearSysid(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(20.0)
    def test_L2_regression(self):
        """Test L2 regression"""
        Ahat = self.notebook_locals["Ahat_2norm"]
        Bhat = self.notebook_locals["Bhat_2norm"]
        X = self.notebook_locals["X"]
        U = self.notebook_locals["U"]
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        self.assertLessEqual(
            np.sum(res**2),
            6e-7,
            "The residual for 2-norm minimization is too high!",
        )

    @weight(3)
    @timeout_decorator.timeout(20.0)
    def test_Linf_regression(self):
        """Test L infinite regression"""
        Ahat = self.notebook_locals["Ahat_infnorm"]
        Bhat = self.notebook_locals["Bhat_infnorm"]
        X = self.notebook_locals["X"]
        U = self.notebook_locals["U"]
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        residual = np.sum(np.max(np.abs(res), axis=0))
        self.assertLessEqual(
            residual,
            0.02,
            "The residual for infinity-norm minimization is too high!",
        )

        obj_infnorm = self.notebook_locals["obj_infnorm"]
        self.assertLessEqual(
            np.abs(obj_infnorm - residual),
            0.1,
            "Objective from optimizer and empirical evaluations are different!",
        )

    @weight(3)
    @timeout_decorator.timeout(20.0)
    def test_L1_regression(self):
        """Test L1 regression"""
        Ahat = self.notebook_locals["Ahat_1norm"]
        Bhat = self.notebook_locals["Bhat_1norm"]
        X = self.notebook_locals["X"]
        U = self.notebook_locals["U"]
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        residual = np.sum(np.abs(res))
        self.assertLessEqual(
            residual, 0.04, "The residual for 1-norm minimization is too high!"
        )

        obj_1norm = self.notebook_locals["obj_1norm"]
        self.assertLessEqual(
            np.abs(obj_1norm - residual),
            0.1,
            "Objective from optimizer and empirical evaluations are different!",
        )



import numpy as np


class Adam:
    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        amsgrad=False,
    ):
        self.params = params
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        self.lr = lr
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        self.eps = eps
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        self.beta1 = betas[0]
        self.beta2 = betas[1]
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        self.weight_decay = weight_decay
        self.amsgrad = amsgrad

        self.exp_avgs = 0 * params
        self.exp_avgs_sqs = 0 * params
        self.max_exp_avgs_sqs = 0 * params
        self.num_steps = 0

    def step(self, loss, dloss_dparams):
        if self.weight_decay != 0.0:
            dloss_dparams += self.weight_decay * params

        self.num_steps += 1

        bias_correction1 = 1 - self.beta1**self.num_steps
        bias_correction2 = 1 - self.beta2**self.num_steps

        self.exp_avgs *= self.beta1
        self.exp_avgs += (1 - self.beta1) * dloss_dparams

        self.exp_avgs_sqs *= self.beta2
        self.exp_avgs_sqs += (1 - self.beta2) * (dloss_dparams**2)

        if self.amsgrad:
            # Maintains the maximum of all 2nd moment running avg. till now
            np.maximumtorch.maximum(
                self.max_exp_avg_sqs, exp_avgs_sqs, out=self.max_exp_avgs_sqs
            )
            # Use the max. for normalizing running avg. of gradient
            denom = (
                np.sqrt(self.max_exp_avgs_sqs) / np.sqrt(bias_correction2)
            ) + self.eps
        else:
            denom = (np.sqrt(self.exp_avgs_sqs) / np.sqrt(bias_correction2)) + self.eps
        step_size = self.lr / bias_correction1
        self.params -= step_size * self.exp_avgs / denom



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight

from underactuated import FindResource


class Testlpdp(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(6)
    @timeout_decorator.timeout(1.0)
    def test_J_from_lp(self):
        """Test optimal cost-to-go from Linear Program"""
        # note: all prints here go to the output item in the json file
        J_value = self.notebook_locals["J_value"]

        J = np.load(FindResource("exercises/dp/J.npy"))

        diff = np.abs(J - J_value)

        self.assertTrue(
            (diff <= 1e-6).all(),
            msg="|J - J*| > 1e-6",
        )



"""
This file contains a number of helper utilities to set up our various
experiments with less code.
"""

import numpy as np
from pydrake.all import (
    BallRpyJoint,
    Box,
    CoulombFriction,
    Cylinder,
    PrismaticJoint,
    RevoluteJoint,
    RigidTransform,
    SpatialInertia,
    Sphere,
    UnitInertia,
)


def AddShape(plant, shape, name, mass=1, mu=1, color=[0.5, 0.5, 0.9, 1.0]):
    instance = plant.AddModelInstance(name)
    # TODO: Add a method to UnitInertia that accepts a geometry shape (unless
    # that dependency is somehow gross) and does this.
    if isinstance(shape, Box):
        inertia = UnitInertia.SolidBox(shape.width(), shape.depth(), shape.height())
    elif isinstance(shape, Cylinder):
        inertia = UnitInertia.SolidCylinder(shape.radius(), shape.length(), [0, 0, 1])
    elif isinstance(shape, Sphere):
        inertia = UnitInertia.SolidSphere(shape.radius())
    elif isinstance(shape, Capsule):
        inertia = UnitInertia.SolidCapsule(shape.radius(), shape.length(), [0, 0, 1])
    else:
        raise RuntimeError(f"need to write the unit inertia for shapes of type {shape}")
    body = plant.AddRigidBody(
        name,
        instance,
        SpatialInertia(mass=mass, p_PScm_E=np.array([0.0, 0.0, 0.0]), G_SP_E=inertia),
    )
    if plant.geometry_source_is_registered():
        if isinstance(shape, Box):
            plant.RegisterCollisionGeometry(
                body,
                RigidTransform(),
                Box(
                    shape.width() - 0.001,
                    shape.depth() - 0.001,
                    shape.height() - 0.001,
                ),
                name,
                CoulombFriction(mu, mu),
            )
            i = 0
            for x in [-shape.width() / 2.0, shape.width() / 2.0]:
                for y in [-shape.depth() / 2.0, shape.depth() / 2.0]:
                    for z in [-shape.height() / 2.0, shape.height() / 2.0]:
                        plant.RegisterCollisionGeometry(
                            body,
                            RigidTransform([x, y, z]),
                            Sphere(radius=1e-7),
                            f"contact_sphere{i}",
                            CoulombFriction(mu, mu),
                        )
                        i += 1
        else:
            plant.RegisterCollisionGeometry(
                body, RigidTransform(), shape, name, CoulombFriction(mu, mu)
            )

        plant.RegisterVisualGeometry(body, RigidTransform(), shape, name, color)

    return instance


# https://github.com/RobotLocomotion/drake/issues/14949
def AddFloatingRpyJoint(plant, frame, instance, use_ball_rpy=True):
    inertia = SpatialInertia(
        mass=0, p_PScm_E=[0.0, 0.0, 0.0], G_SP_E=UnitInertia(0, 0, 0)
    )
    x_body = plant.AddRigidBody("x", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("x", plant.world_frame(), x_body.body_frame(), [1, 0, 0])
    )
    y_body = plant.AddRigidBody("y", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("y", x_body.body_frame(), y_body.body_frame(), [0, 1, 0])
    )
    z_body = plant.AddRigidBody("z", instance, inertia)
    plant.AddJoint(
        PrismaticJoint("z", y_body.body_frame(), z_body.body_frame(), [0, 0, 1])
    )
    if use_ball_rpy:
        plant.AddJoint(BallRpyJoint("ball", z_body.body_frame(), frame))
    else:
        # RollPitchYaw is body z-y-x
        rz_body = plant.AddRigidBody("rz", instance, inertia)
        plant.AddJoint(
            RevoluteJoint("rz", z_body.body_frame(), rz_body.body_frame(), [0, 0, 1])
        )
        ry_body = plant.AddRigidBody("ry", instance, inertia)
        plant.AddJoint(
            RevoluteJoint("ry", rz_body.body_frame(), ry_body.body_frame(), [0, 1, 0])
        )
        plant.AddJoint(RevoluteJoint("rx", ry_body.body_frame(), frame, [1, 0, 0]))



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestIlqrDriving(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test1_discrete_dynamics(self):
        """Test discrete_dynamics"""
        # load locals
        f = self.notebook_locals["discrete_dynamics"]

        f_target = [
            np.array([0.14181286, 0.81062987, 0.54580749, 0.77731476, 1.02810156]),
            np.array([0.131663, 0.300996, 0.57034388, 0.71732411, 0.81033267]),
            np.array([0.33233256, 0.91916814, 0.27416228, 0.45461388, 0.99126091]),
            np.array([1.02770659, 0.27770476, 0.56066881, 0.96146963, 0.20821043]),
            np.array([0.71706372, 0.47773599, 0.22401885, 0.538506, 0.40897373]),
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            x = np.random.rand(5)
            u = np.random.rand(2)
            f_eval = f(x, u)
            np.testing.assert_array_almost_equal(
                f_eval,
                f_target[i],
                err_msg="The discrete dynamics are incorrect. "
                f"Expected output {f_target[i]} with x={x} and u={u}",
            )

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test2_rollout(self):
        """Test rollout"""
        # load locals
        f = self.notebook_locals["rollout"]

        f_target_test_target = [
            0.36531949939751773,
            2.25991775105549,
            -0.05889795147375571,
            0.5476245815855634,
            0.8066768897676444,
        ]

        np.random.seed(7)
        n_rands = 5
        for i in range(n_rands):
            N = 5
            n_u = 2
            x0 = np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            f_eval_test = np.random.randn(N) @ f(x0, u_trj) @ np.random.randn(N)
            msg = (
                f"The discrete dynamics are incorrect."
                + "Expected output {f_target_test_target[i]} got {f_eval_test}"
            )
            self.assertAlmostEqual(f_eval_test, f_target_test_target[i], msg=msg)

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test3_cost_trj(self):
        """Test cost_trj"""
        # load locals
        f = self.notebook_locals["cost_trj"]

        f_target = [
            29.946731429035616,
            34.805336401404986,
            27.626872505986434,
            33.25730851334137,
            33.83897480138296,
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            N = 5
            n_u = 2
            n_x = 5
            np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            x_trj = np.random.randn(N, n_x) * 0.4
            f_eval = f(x_trj, u_trj)
            np.testing.assert_almost_equal(
                f_eval,
                f_target[i],
                err_msg="The cost_trj function is incorrect. "
                f"Expected output {f_target[i]} with x_trj=\n{x_trj}"
                f" and u=\n{u_trj}",
            )

    @weight(5)
    @timeout_decorator.timeout(1.0)
    def test4_Q_terms(self):
        """Test Q_terms"""
        # load locals
        Q_terms = self.notebook_locals["Q_terms"]
        derivs = self.notebook_locals["derivs"]

        np.random.seed(7)
        x = np.random.rand(5)
        u = np.random.rand(2)

        l_x, l_u, l_xx, l_ux, l_uu, f_x, f_u = derivs.stage(x, u)
        V_x, V_xx = derivs.final(x + 0.1)
        Q_x, Q_u, Q_xx, Q_ux, Q_uu = Q_terms(
            l_x, l_u, l_xx, l_ux, l_uu, f_x, f_u, V_x, V_xx
        )

        Q_x_test_val = Q_x @ np.random.randn(Q_x.shape[0])
        Q_u_test_val = Q_u @ np.random.randn(Q_u.shape[0])
        Q_xx_test_val = (
            np.random.randn(Q_xx.shape[0]) @ Q_xx @ np.random.randn(Q_xx.shape[0])
        )
        Q_ux_test_val = (
            np.random.randn(Q_ux.shape[0]) @ Q_ux @ np.random.randn(Q_ux.shape[1])
        )
        Q_uu_test_val = (
            np.random.randn(Q_uu.shape[0]) @ Q_uu @ np.random.randn(Q_uu.shape[0])
        )

        Q_x_test_target = 1.1817013977437874
        Q_u_test_target = -0.1716142906672626
        Q_xx_test_target = 0.5337352394014466
        Q_ux_test_target = 0.5156605288974515
        Q_uu_test_target = -0.28166353613679074

        self.assertAlmostEqual(Q_x_test_val, Q_x_test_target, msg="Incorrect Q_x")
        self.assertAlmostEqual(Q_u_test_val, Q_u_test_target, msg="Incorrect Q_u")
        self.assertAlmostEqual(Q_xx_test_val, Q_xx_test_target, msg="Incorrect Q_xx")
        self.assertAlmostEqual(Q_ux_test_val, Q_ux_test_target, msg="Incorrect Q_ux")
        self.assertAlmostEqual(Q_uu_test_val, Q_uu_test_target, msg="Incorrect Q_uu")

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test5_gains(self):
        """Test gains"""
        # load locals
        gains = self.notebook_locals["gains"]

        np.random.seed(8)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        k, K = gains(Q_uu, Q_u, Q_ux)

        test_inputs = 10 * np.random.randn(k.shape[0], 5)

        k_test_val = (k @ test_inputs).sum()
        K_test_val = (test_inputs.T @ K).sum()

        self.assertAlmostEqual(k_test_val, 10.712481345027399, msg="Incorrect k gain")

        self.assertAlmostEqual(K_test_val, -12.949530490536842, msg="Incorrect K gain")

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test6_V_terms(self):
        """Test V_terms"""
        # load locals
        V_terms = self.notebook_locals["V_terms"]
        gains = self.notebook_locals["gains"]

        np.random.seed(7)
        Q_x = np.random.randn(5)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        Q_xx = np.random.randn(5, 5)
        Q_xx = 0.5 * (Q_xx + Q_xx.T)
        k, K = gains(Q_uu, Q_u, Q_ux)

        V_x, V_xx = V_terms(Q_x, Q_u, Q_xx, Q_ux, Q_uu, K, k)
        test_inputs = np.random.randn(V_x.shape[0])

        V_x_test_val = V_x @ test_inputs
        V_xx_test_val = test_inputs @ V_xx @ test_inputs

        self.assertAlmostEqual(V_x_test_val, -2.113959925580339, msg="Incorrect V_x")

        self.assertAlmostEqual(V_xx_test_val, 35.931091753768, msg="Incorrect V_xx")

        # this tests whether the student simplified the V_terms expression
        k2, K2 = 10 * np.random.randn(*k.shape), 10 * np.random.randn(*K.shape)
        V_x2, V_xx2 = V_terms(Q_x, Q_u, Q_xx, Q_ux, Q_uu, K2, k2)

        V_x2_test_val = V_x2 @ test_inputs
        V_xx2_test_val = test_inputs @ V_xx2 @ test_inputs

        self.assertNotAlmostEqual(
            V_x2_test_val,
            2.484304411129088,
            msg="""It appears you may have simplified the expression for V_x.
            Do not do this""",
        )
        self.assertNotAlmostEqual(
            V_xx2_test_val,
            265.0164583112695,
            msg="""It appears you may have simplified the expression for V_xx.
            Do not do this""",
        )

        self.assertAlmostEqual(V_x2_test_val, -6.043415622930539, msg="Incorrect V_x")

        self.assertAlmostEqual(V_xx2_test_val, -52.01732694080634, msg="Incorrect V_xx")

    @weight(2)
    @timeout_decorator.timeout(1.0)
    def test7_forward_pass(self):
        """Test forward_pass"""
        # load locals
        forward_pass = self.notebook_locals["forward_pass"]

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)
        k_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1])
        K_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1], x_trj.shape[1])
        x_trj_new, u_trj_new = forward_pass(x_trj, u_trj, k_trj, K_trj)

        x_test = np.random.randn(x_trj_new.shape[0])
        u_test = np.random.randn(u_trj_new.shape[0])

        x_trj_new_test_val = x_test @ x_trj_new @ x_test
        u_trj_new_test_val = (u_test @ u_trj_new).sum()

        self.assertAlmostEqual(
            x_trj_new_test_val, -1.0615491271322297, msg="Incorrect x_trj_new"
        )

        self.assertAlmostEqual(
            u_trj_new_test_val, -10.649763284585609, msg="Incorrect u_trj_new"
        )

    @weight(2)
    @timeout_decorator.timeout(1.0)
    def test8_backward_pass(self):
        """Test backward_pass"""
        # load locals
        backward_pass = self.notebook_locals["backward_pass"]

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        np.random.randn(2)
        np.random.randn(2, 5)
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)

        k_trj, K_trj, _ = backward_pass(x_trj, u_trj, 100.0)

        r_test = 10 * np.random.randn(k_trj.shape[1])
        l_test = 10 * np.random.randn(k_trj.shape[0])

        k_trj_test_val = l_test @ k_trj @ r_test
        K_trj_test_val = np.einsum("i, ijk, j -> k", l_test, K_trj, r_test).sum()

        self.assertAlmostEqual(
            k_trj_test_val, -5.380572857461254, msg="Incorrect k_trj"
        )

        self.assertAlmostEqual(K_trj_test_val, 3.770894929116537, msg="Incorrect K_trj")



import os
from urllib.request import urlretrieve

from pydrake.all import PackageMap, namedview

running_as_test = False


def _set_running_as_test(value):
    """[INTERNAL USE ONLY]: Set the global variable `running_as_test` to
    `value`.

    This method is used by the build system; it is not intended for general
    use.
    """
    global running_as_test
    running_as_test = value


def FindResource(filename):
    """Returns the absolute path to the given filename relative to the
    underactuated module."""
    return os.path.join(os.path.dirname(__file__), filename)


def FindDataResource(filename: str):
    """
    Returns the absolute path to the given filename relative to the data directory; fetching it from a remote host if necessary.
    """
    data = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data")
    if not os.path.exists(data):
        os.makedirs(data)
    path = os.path.join(data, filename)
    if not os.path.exists(path):
        print(f"{path} was not found locally; downloading it now...")
        urlretrieve(f"https://underactuated.csail.mit.edu/data/{filename}", path)
    return path


def ConfigureParser(parser):
    """Add the underactuated module packages to the given Parser."""
    package_xml = os.path.join(os.path.dirname(__file__), "package.xml")
    parser.package_map().AddPackageXml(filename=package_xml)
    # Add spot_description
    parser.package_map().AddRemote(
        package_name="spot_description",
        params=PackageMap.RemoteParams(
            urls=[
                f"https://github.com/bdaiinstitute/spot_ros2/archive/097f46bd7c5b1d6bf0189a895c28ae0a90d287b1.tar.gz"
            ],
            sha256=("59df2f0644bd7d937365e580a6d3b69822e7fb0358dc58456cd0408830641d88"),
            strip_prefix="spot_ros2-097f46bd7c5b1d6bf0189a895c28ae0a90d287b1/spot_description/",
        ),
    )


def Rgba2Hex(rgb):
    """
    Turn a list of R,G,B elements (any indexable list of >= 3 elements will
    work), where each element is specified on range [0., 1.], into the
    equivalent 24-bit value 0xRRGGBB.
    """
    val = 0
    for i in range(3):
        val += (256 ** (2 - i)) * int(255 * rgb[i])
    return val


def MakeNamedViewPositions(mbp, view_name, add_suffix_if_single_position=False):
    print(
        "MakeNamedViewPositions() is deprecated.  Use `namedview(view_name, plant.GetPositionNames())` instead)"
    )
    return namedview(
        view_name,
        plant.GetPositionNames(
            add_model_instance_prefix=False,
            always_add_suffix=not add_suffix_if_single_position,
        ),
    )


def MakeNamedViewVelocities(mbp, view_name, add_suffix_if_single_velocity=False):
    print(
        "MakeNamedViewVelocities() is deprecated.  Use `namedview(view_name, plant.GetVelocityNames())` instead)"
    )
    return namedview(
        view_name,
        plant.GetVelocityNames(
            add_model_instance_prefix=False,
            always_add_suffix=not add_suffix_if_single_velocity,
        ),
    )


def MakeNamedViewState(mbp, view_name):
    print(
        "MakeNamedViewState() is deprecated.  Use `namedview(view_name, plant.GetStateNames())` instead)"
    )
    return namedview(
        view_name,
        plant.GetStateNames(add_model_instance_prefix=False),
    )


def MakeNamedViewActuation(mbp, view_name):
    print(
        "MakeNamedViewActuation() is deprecated.  Use `namedview(view_name, plant.GetActuatorNames())` instead)"
    )
    return namedview(
        view_name,
        plant.GetActuatorNames(add_model_instance_prefix=False),
    )



import time
import typing
from functools import partial
from inspect import Parameter, signature

import numpy as np
from ipywidgets.widgets.interaction import (
    _get_min_max_value,
    _yield_abbreviations_for_parameter,
)
from pydrake.geometry import Cylinder, Meshcat, MeshcatVisualizer, Rgba, Sphere
from pydrake.math import RigidTransform, RotationMatrix
from pydrake.multibody.plant import MultibodyPlant
from pydrake.solvers import (
    BoundingBoxConstraint,
    MathematicalProgram,
    MathematicalProgramResult,
)
from pydrake.systems.framework import Context, EventStatus, LeafSystem
from pydrake.trajectories import Trajectory

from underactuated import running_as_notebook


# This class is scheduled for removal.  Use meshcat interaction methods instead.
def _interact(meshcat, callback, **kwargs):
    """
    A poor-person's implementation of a functionality like ipywidgets interact
    https://ipywidgets.readthedocs.io/en/latest/examples/Using%20Interact.html#Basic-interact
    """
    values = {}

    # The following code was adapted from https://github.com/jupyter-widgets/ipywidgets/blob/9326d28c474cb6d2b58876d469489a73e98f903a/python/ipywidgets/ipywidgets/widgets/interaction.py  # noqa
    def find_abbreviations(callback, kwargs):
        """
        Find the abbreviations for the given function and kwargs.
        Return (name, abbrev, default) tuples.
        """
        new_kwargs = []
        try:
            sig = signature(callback)
        except (ValueError, TypeError):
            # can't inspect, no info from function; only use kwargs
            return [(key, value, value) for key, value in kwargs.items()]

        for param in sig.parameters.values():
            for name, value, default in _yield_abbreviations_for_parameter(
                param, kwargs
            ):
                if value is Parameter.empty:
                    raise ValueError(
                        "cannot find widget or abbreviation for argument: {!r}".format(
                            name
                        )
                    )
                new_kwargs.append((name, value, default))
        return new_kwargs

    new_kwargs = find_abbreviations(callback, kwargs)

    for name, abbrev, default in new_kwargs:
        kw = {}
        kw["value"] = None if default is Parameter.empty else default
        if isinstance(abbrev, tuple):
            kw["step"] = abbrev[2] if len(abbrev) != 3 else None
            kw["min"], kw["max"], kw["value"] = _get_min_max_value(
                abbrev[0], abbrev[1], **kw
            )
            if kw["step"] is None:
                kw["step"] = 0.1
            meshcat.AddSlider(name, **kw)
        else:
            raise ValueError("This case is not implemented yet")
            # It might be simple.  I just haven't tried!
        values[name] = kw["value"]

    def update_values():
        changed = False
        for name in values:
            v = meshcat.GetSliderValue(name)
            changed |= v != values[name]
            values[name] = v
        return changed

    # Always call it once.
    callback(**values)

    if not running_as_notebook:
        return

    print("Press the 'Stop Interacting' button in Meshcat to continue.")
    meshcat.AddButton("Stop Interacting")
    while meshcat.GetButtonClicks("Stop Interacting") < 1:
        if update_values():
            callback(**values)
        time.sleep(0.1)

    meshcat.DeleteButton("Stop Interacting")
    for name in values:
        meshcat.DeleteSlider(name)


class MeshcatSliders(LeafSystem):
    """
    A system that outputs the values from meshcat sliders.

    An output port is created for each element in the list `slider_names`.
    Each element of `slider_names` must itself be an iterable collection
    (list, tuple, set, ...) of strings, with the names of sliders that have
    *already* been added to Meshcat via Meshcat.AddSlider().

    The same slider may be used in multiple ports.
    """

    def __init__(self, meshcat: Meshcat, slider_names: typing.List[str]):
        LeafSystem.__init__(self)

        self._meshcat = meshcat
        self._sliders = slider_names
        for i, slider_iterable in enumerate(self._sliders):
            port = self.DeclareVectorOutputPort(
                f"slider_group_{i}",
                len(slider_iterable),
                partial(self._DoCalcOutput, port_index=i),
            )
            port.disable_caching_by_default()

    def _DoCalcOutput(self, context, output, port_index):
        for i, slider in enumerate(self._sliders[port_index]):
            output[i] = self._meshcat.GetSliderValue(slider)


class WsgButton(LeafSystem):
    """Adds a button named `Open/Close Gripper` to the meshcat GUI, and registers the Space key to press it. Pressing this button will toggle the value of the output port from a wsg position command corresponding to an open position or a closed position.

    Args:
        meshcat: The meshcat instance in which to register the button.
    """

    def __init__(self, meshcat: Meshcat):
        LeafSystem.__init__(self)
        port = self.DeclareVectorOutputPort("wsg_position", 1, self._DoCalcOutput)
        port.disable_caching_by_default()
        self._meshcat = meshcat
        self._button = "Open/Close Gripper"
        meshcat.AddButton(self._button, "Space")
        print("Press Space to open/close the gripper")

    def __del__(self):
        self._meshcat.DeleteButton(self._button)

    def _DoCalcOutput(self, context, output):
        position = 0.107  # open
        if (self._meshcat.GetButtonClicks(self._button) % 2) == 1:
            position = 0.002  # close
        output.SetAtIndex(0, position)


class StopButton(LeafSystem):
    """Adds a button named `Stop Simulation` to the meshcat GUI, and registers
    the `Escape` key to press it. Pressing this button will terminate the
    simulation.

    Args:
        meshcat: The meshcat instance in which to register the button.
        check_interval: The period at which to check for button presses.
    """

    def __init__(self, meshcat: Meshcat, check_interval: float = 0.1):
        LeafSystem.__init__(self)
        self._meshcat = meshcat
        self._button = "Stop Simulation"

        self.DeclareDiscreteState([0])  # button click count
        self.DeclareInitializationDiscreteUpdateEvent(self._Initialize)
        self.DeclarePeriodicDiscreteUpdateEvent(check_interval, 0, self._CheckButton)

        # Create the button now (rather than at initialization) so that the
        # CheckButton method will work even if Initialize has never been
        # called.
        meshcat.AddButton(self._button, "Escape")

    def __del__(self):
        # TODO(russt): Provide a nicer way to check if the button is currently
        # registered.
        try:
            self._meshcat.DeleteButton(self._button)
        except:
            pass

    def _Initialize(self, context, discrete_state):
        print("Press Escape to stop the simulation")
        discrete_state.set_value([self._meshcat.GetButtonClicks(self._button)])

    def _CheckButton(self, context, discrete_state):
        clicks_at_initialization = context.get_discrete_state().value()[0]
        if self._meshcat.GetButtonClicks(self._button) > clicks_at_initialization:
            self._meshcat.DeleteButton(self._button)
            return EventStatus.ReachedTermination(self, "Termination requested by user")
        return EventStatus.DidNothing()


def AddMeshcatTriad(
    meshcat: Meshcat,
    path: str,
    length: float = 0.25,
    radius: float = 0.01,
    opacity: float = 1.0,
    X_PT: RigidTransform = RigidTransform(),
):
    """Adds an X-Y-Z triad to the meshcat scene.

    Args:
        meshcat: A Meshcat instance.
        path: The Meshcat path on which to attach the triad. Using relative paths will attach the triad to the path's coordinate system.
        length: The length of the axes in meters.
        radius: The radius of the axes in meters.
        opacity: The opacity of the axes in [0, 1].
        X_PT: The pose of the triad relative to the path.
    """
    meshcat.SetTransform(path, X_PT)
    # x-axis
    X_TG = RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0])
    meshcat.SetTransform(path + "/x-axis", X_TG)
    meshcat.SetObject(
        path + "/x-axis", Cylinder(radius, length), Rgba(1, 0, 0, opacity)
    )

    # y-axis
    X_TG = RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0])
    meshcat.SetTransform(path + "/y-axis", X_TG)
    meshcat.SetObject(
        path + "/y-axis", Cylinder(radius, length), Rgba(0, 1, 0, opacity)
    )

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.0])
    meshcat.SetTransform(path + "/z-axis", X_TG)
    meshcat.SetObject(
        path + "/z-axis", Cylinder(radius, length), Rgba(0, 0, 1, opacity)
    )


def plot_mathematical_program(
    meshcat: Meshcat,
    path: str,
    prog: MathematicalProgram,
    X: np.ndarray,
    Y: np.ndarray,
    result: MathematicalProgramResult = None,
    point_size: float = 0.05,
):
    """Visualize a MathematicalProgram in Meshcat.

    Args:
        meshcat: A Meshcat instance.
        path: The Meshcat path to the visualization.
        prog: A MathematicalProgram instance.
        X: A 2D array of x values.
        Y: A 2D array of y values.
        result: A MathematicalProgramResult instance; if provided then the
            solution is visualized as a point.
        point_size: The size of the solution point.
    """
    assert prog.num_vars() == 2
    assert X.size == Y.size

    X.size
    values = np.vstack((X.reshape(-1), Y.reshape(-1)))
    costs = prog.GetAllCosts()

    # Vectorized multiply for the quadratic form.
    # Z = (D*np.matmul(Q,D)).sum(0).reshape(nx, ny)

    if costs:
        Z = prog.EvalBindingVectorized(costs[0], values)
        for b in costs[1:]:
            Z = Z + prog.EvalBindingVectorized(b, values)

    cv = f"{path}/constraints"
    for binding in prog.GetAllConstraints():
        if isinstance(binding.evaluator(), BoundingBoxConstraint):
            c = binding.evaluator()
            var_indices = [
                int(prog.decision_variable_index()[v.get_id()])
                for v in binding.variables()
            ]
            satisfied = np.array(
                c.CheckSatisfiedVectorized(values[var_indices, :], 0.001)
            ).reshape(1, -1)
            if costs:
                Z[~satisfied] = np.nan

            v = f"{cv}/{type(c).__name__}"
            Zc = np.zeros(Z.shape)
            Zc[satisfied] = np.nan
            meshcat.PlotSurface(
                v,
                X,
                Y,
                Zc.reshape(X.shape),
                rgba=Rgba(1.0, 0.2, 0.2, 1.0),
                wireframe=True,
            )
        else:
            Zc = prog.EvalBindingVectorized(binding, values)
            evaluator = binding.evaluator()
            low = evaluator.lower_bound()
            up = evaluator.upper_bound()
            cvb = f"{cv}/{type(evaluator).__name__}"
            for index in range(Zc.shape[0]):
                # TODO(russt): Plot infeasible points in a different color.
                infeasible = np.logical_or(
                    Zc[index, :] < low[index], Zc[index, :] > up[index]
                )
                meshcat.PlotSurface(
                    f"{cvb}/{index}",
                    X,
                    Y,
                    Zc[index, :].reshape(X.shape),
                    rgba=Rgba(1.0, 0.3, 1.0, 1.0),
                    wireframe=True,
                )

    if costs:
        meshcat.PlotSurface(
            f"{path}/objective",
            X,
            Y,
            Z.reshape(X.shape),
            rgba=Rgba(0.3, 1.0, 0.3, 1.0),
            wireframe=True,
        )

    if result:
        v = f"{path}/solution"
        meshcat.SetObject(v, Sphere(point_size), Rgba(0.3, 1.0, 0.3, 1.0))
        x_solution = result.get_x_val()
        meshcat.SetTransform(
            v,
            RigidTransform([x_solution[0], x_solution[1], result.get_optimal_cost()]),
        )


def PublishPositionTrajectory(
    trajectory: Trajectory,
    root_context: Context,
    plant: MultibodyPlant,
    visualizer: MeshcatVisualizer,
    time_step: float = 1.0 / 33.0,
):
    """
    Publishes an animation to Meshcat of a MultibodyPlant using a trajectory of the plant positions.

    Args:
        trajectory: A Trajectory instance.
        root_context: The root context of the diagram containing plant.
        plant: A MultibodyPlant instance.
        visualizer: A MeshcatVisualizer instance.
        time_step: The time step between published frames.
    """
    plant_context = plant.GetMyContextFromRoot(root_context)
    visualizer_context = visualizer.GetMyContextFromRoot(root_context)

    visualizer.StartRecording(False)

    for t in np.append(
        np.arange(trajectory.start_time(), trajectory.end_time(), time_step),
        trajectory.end_time(),
    ):
        root_context.SetTime(t)
        plant.SetPositions(plant_context, trajectory.value(t)[: plant.num_positions()])
        visualizer.ForcedPublish(visualizer_context)

    visualizer.StopRecording()
    visualizer.PublishRecording()



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestRRTPlanning(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(7)
    @timeout_decorator.timeout(1.0)
    def test1_rrt_path(self):
        """Test RRT"""
        # load locals
        RRT = self.notebook_locals["RRT"]

        start = np.array([11, 0])
        goal = np.array([6, 8])

        obstacles = [
            np.array([9, 6, 2]),
            np.array([9, 8, 1]),
            np.array([9, 10, 2]),
            np.array([4, 5, 2]),
            np.array([7, 5, 2]),
            np.array([4, 10, 1]),
        ]

        bounds = np.array([-2, 15])

        np.random.seed(7)
        rrt = RRT(start=start, goal=goal, bounds=bounds, obstacle_list=obstacles)
        path = rrt.plan()

        self.assertFalse(path is None, "The plan method is not implemented correctly")

        np.testing.assert_array_almost_equal(
            start,
            path[-1],
            err_msg="The start location of the path is not the correct "
            + "start location",
        )

        np.testing.assert_array_almost_equal(
            goal,
            path[0],
            err_msg="The goal location of the path is not the correct "
            + "goal location",
        )

        for i in range(len(path) - 1):
            is_in_collision = RRT.collision(
                RRT.Node(path[i]), RRT.Node(path[i + 1]), obstacles
            )
            self.assertFalse(is_in_collision, "The path is colliding with obstacles")

    @weight(7)
    @timeout_decorator.timeout(20.0)
    def test2_rrt_star_path(self):
        """Test RRT*"""
        # load locals
        RRTStar = self.notebook_locals["RRTStar"]

        start = np.array([11, 0])
        goal = np.array([6, 8])

        obstacles = [
            np.array([9, 6, 2]),
            np.array([9, 8, 1]),
            np.array([9, 10, 2]),
            np.array([4, 5, 2]),
            np.array([7, 5, 2]),
            np.array([4, 10, 1]),
        ]

        bounds = np.array([-2, 15])

        rrt_star = RRTStar(
            start=start, goal=goal, bounds=bounds, obstacle_list=obstacles
        )

        # Check the choose_parent method
        node1 = rrt_star.Node(np.array([12, 0]))
        node1 = rrt_star.steer(rrt_star.start, node1)
        node1.cost = 10
        node2 = rrt_star.Node(np.array([10, 0]))
        node2 = rrt_star.steer(rrt_star.start, node2)
        node2.cost = 20
        node3 = rrt_star.Node(np.array([5, 8]))
        node3 = rrt_star.steer(rrt_star.start, node3)
        node3.cost = 0
        node4 = rrt_star.Node(np.array([5, 0]))
        rrt_star.node_list = [node1, node2, node3]
        node4 = rrt_star.steer(rrt_star.start, node4)
        near_inds = rrt_star.near_nodes_inds(node4)
        node4 = rrt_star.choose_parent(node4, near_inds)

        self.assertFalse(
            node4.parent is None,
            "The choose_parent method is not implemented correctly",
        )

        self.assertFalse(
            np.linalg.norm(node3.p - node4.parent.p) < 1e-5,
            "The choose_parent method does not check for collisions",
        )

        np.testing.assert_array_almost_equal(
            node4.parent.p,
            node1.p,
            err_msg="The choose_parent method does not seem to"
            + " select the best parent node",
        )

        # Check the rewiring method
        node4 = rrt_star.steer(rrt_star.start, node4)
        rrt_star.rewire(node4, near_inds)
        np.testing.assert_array_almost_equal(
            node1.parent.p,
            node4.p,
            err_msg="The rewiring method does not rewire the correct nodes",
        )

        node3.cost = 100
        node4 = rrt_star.steer(rrt_star.start, node4)
        rrt_star.rewire(node4, near_inds)
        self.assertFalse(
            np.linalg.norm(node3.parent.p - node4.p) < 1e-5,
            "The rewiring method does not check for collision",
        )

        np.random.seed(7)
        rrt_star = RRTStar(
            start=start,
            goal=goal,
            bounds=bounds,
            obstacle_list=obstacles,
            goal_sample_rate=0.0,
            max_iter=300,
        )
        path, _ = rrt_star.plan()

        np.testing.assert_array_almost_equal(
            start,
            path[-1],
            err_msg="The start location of the path is not the correct "
            + "start location",
        )

        np.testing.assert_array_almost_equal(
            goal,
            path[0],
            err_msg="The goal location of the path is not the correct "
            + "goal location",
        )

        for i in range(len(path) - 1):
            is_in_collision = RRTStar.collision(
                RRTStar.Node(path[i]), RRTStar.Node(path[i + 1]), obstacles
            )
            self.assertFalse(is_in_collision, "The path is colliding with obstacles")

        def path_cost(path):
            total = 0
            for i in range(len(path) - 1):
                total += np.linalg.norm(path[i] - path[i + 1])
            return total

        path_length = path_cost(path)
        self.assertTrue(
            path_length < 17.5,
            "Length of the path={} > 17.5".format(path_length),
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import AffineSystem, Diagram, System


class TestDrakeDiagrams(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_actuator_model(self):
        # note: all prints here go to the output item in the json file

        actuator_model = self.notebook_locals["actuator_model"]

        assert isinstance(actuator_model, System), "actuator_model must be a System"
        assert (
            actuator_model.ToAutoDiffXdMaybe()
        ), "Your system isn't compatible with Drake's AutoDiffXd; see the note above."  # noqa
        assert actuator_model.num_input_ports() == 1, "must have exactly one input port"
        assert (
            actuator_model.get_input_port().size() == 1
        ), "input port must have size 1"
        assert actuator_model.num_output_ports() == 1, "must have only one output port"
        assert (
            actuator_model.get_output_port().size() == 1
        ), "output port should have size 1"
        assert (
            actuator_model.num_continuous_states() == 1
        ), "should have exactly one continuous state variable"
        assert (
            actuator_model.num_discrete_state_groups() == 0
        ), "should not have any discrete state"

        context = actuator_model.CreateDefaultContext()

        def check_dynamics(x, u):
            context.SetContinuousState([x])
            actuator_model.get_input_port().FixValue(context, [u])
            xdot = actuator_model.EvalTimeDerivatives(context).CopyToVector()
            self.assertAlmostEqual(
                2 * (u - x),
                xdot[0],
                "the state dynamics (xdot) of your actuator model are incorrect",
            )
            y = actuator_model.get_output_port().Eval(context)
            self.assertAlmostEqual(
                y, x, "the output value of your actuator model is incorrect"
            )

        check_dynamics(x=0, u=0)
        check_dynamics(x=0.4, u=0.1)
        check_dynamics(x=0.52, u=0.25)
        check_dynamics(x=-0.91, u=5.2)
        check_dynamics(x=2, u=-2)

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_diagram(self):
        diagram = self.notebook_locals["diagram"]
        assert isinstance(diagram, Diagram), "diagram must be a Diagram"
        assert (
            len(diagram.GetSystems()) == 2
        ), "diagram should have exactly two subsystems"  # noqa
        assert (
            diagram.ToAutoDiffXdMaybe()
        ), "Your system isn't compatible with Drake's AutoDiffXd; see the note above."  # noqa
        assert (
            diagram.num_input_ports() == 1
        ), "diagram should have exactly one input port"
        assert (
            diagram.get_input_port().size() == 1
        ), "diagram's input port should have size 1"
        assert (
            diagram.num_continuous_states() == 3
        ), "diagram should have a total of 3 continuous states"
        assert (
            diagram.num_discrete_state_groups() == 0
        ), "diagram should not have any discrete states"

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_diagram_context(self):
        diagram = self.notebook_locals["diagram"]
        double_integrator = self.notebook_locals["double_integrator"]
        actuator_model = self.notebook_locals["actuator_model"]
        context = self.notebook_locals["diagram_context"]

        double_integrator_context = double_integrator.GetMyContextFromRoot(context)
        np.testing.assert_almost_equal(
            double_integrator_context.get_continuous_state_vector().get_value(),
            [1.2, 0],
            err_msg="the value of the double integrator state is incorrect",
        )
        actuator_model_context = actuator_model.GetMyContextFromRoot(context)
        np.testing.assert_almost_equal(
            actuator_model_context.get_continuous_state_vector().get_value(),
            [0],
            err_msg="the value of the actuator model state is incorrect",
        )

        assert diagram.get_input_port().HasValue(
            context
        ), "you must fix a value for the diagram's input port"
        np.testing.assert_almost_equal(
            diagram.get_input_port().Eval(context),
            [0],
            err_msg="the input port value is incorrect",
        )

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_controller(self):
        controller = self.notebook_locals["controller"]
        assert isinstance(
            controller, AffineSystem
        ), "controller should be an affine system"

        assert (
            controller.num_continuous_states() == 0
        ), "your controller has state; that shouldn't happen"
        np.testing.assert_almost_equal(
            np.sort(controller.D()),
            np.sort([[-1.06519664, -1.0, -2.26503715]]),
            err_msg="your LQR gains do not match the expected value",
        )



import numpy as np
from IPython.display import HTML, display
from matplotlib.animation import HTMLWriter
from matplotlib.widgets import Slider
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Context, VectorSystem
from pydrake.systems.pyplot_visualizer import PyPlotVisualizer
from pydrake.trajectories import Trajectory

from underactuated.jupyter import running_as_notebook


class HistogramVisualizer(PyPlotVisualizer):
    """A simple visualizer that plots a histogram of the input vector."""

    def __init__(
        self, num_samples, bins, xlim, ylim, draw_time_step, figsize=(6, 4), show=True
    ):
        PyPlotVisualizer.__init__(self, draw_time_step, figsize=figsize, show=show)
        self.DeclareVectorInputPort(f"x", num_samples)
        self.num_samples = num_samples
        self.bins = bins
        self.data = [0] * num_samples
        self.scale = 10
        self.limits = xlim
        self.ax.set_xlim(xlim)
        self.ax.axis("auto")
        self.ax.set_ylim(ylim)
        self.patches = None

    def draw(self, context):
        if self.patches:
            [p.remove() for p in self.patches]
        self.data = self.EvalVectorInput(context, 0).value()
        count, bins, self.patches = self.ax.hist(
            self.data,
            bins=self.bins,
            range=self.limits,
            density=False,
            weights=[self.scale / self.num_samples] * self.num_samples,
            facecolor="b",
        )
        self.ax.set_title("t = " + str(context.get_time()))


class SliderSystem(VectorSystem):
    def __init__(self, ax, title, min, max):
        # 0 inputs, 1 output.
        VectorSystem.__init__(self, 0, 1)
        self.value = 0
        self.slider = Slider(ax, title, min, max, valinit=self.value)
        self.slider.on_changed(self.update)

    def update(self, val):
        self.value = val

    def DoCalcVectorOutput(self, context, unused, unused2, output):
        output[:] = self.value


def AdvanceToAndVisualize(
    simulator,
    visualizer,
    time,
    time_if_running_headless=0.1,
    movie_filename=None,
):
    """
    Helper to support visualizing a simulation with pyplot visualizer.
    Will simply simulate (with target_realtime_rate = 1) if visualizer.show =
    True, or will record and render an animation if visualizer.show = False. If
    specified, time_if_running_headless will be used instead of time if
    running_as_notebook is False.
    """
    if visualizer._show:
        target_rate = simulator.get_target_realtime_rate()
        simulator.set_target_realtime_rate(1.0)
    else:
        print("simulating... ", end=" ")
        visualizer.start_recording()
    if time_if_running_headless and not running_as_notebook:
        time = time_if_running_headless
    simulator.AdvanceTo(time)
    if not visualizer._show or movie_filename:
        print("done.\ngenerating animation...")
        ani = visualizer.get_recording_as_animation()
        display(HTML(ani.to_jshtml()))
        if movie_filename:
            with open(movie_filename, "w") as f:
                f.write(ani.to_jshtml())
    else:
        simulator.set_target_realtime_rate(target_rate)


def AdvanceToAndSaveAnimation(simulator, visualizer, time, filename):
    visualizer.start_recording()
    simulator.AdvanceTo(time)
    ani = visualizer.get_recording_as_animation()
    # Note: Wanted to use embed_frames=True, but it did not render
    # the image properly for me.
    writer = HTMLWriter()
    writer.frame_format = "svg"
    ani.save(filename, writer=writer)


def AnimatePositionTrajectory(
    trajectory: Trajectory,
    root_context: Context,
    plant: MultibodyPlant,
    visualizer: PyPlotVisualizer,
    time_step: float = 1.0 / 33.0,
):
    """
    Returns an animation of a MultibodyPlant using a trajectory of the plant positions.

    Args:
        trajectory: A Trajectory instance.
        root_context: The root context of the diagram containing plant.
        plant: A MultibodyPlant instance.
        visualizer: A PyPlotVisualizer instance.
        time_step: The time step between published frames.
    """
    plant_context = plant.GetMyContextFromRoot(root_context)
    visualizer_context = visualizer.GetMyContextFromRoot(root_context)

    visualizer.start_recording()

    for t in np.append(
        np.arange(trajectory.start_time(), trajectory.end_time(), time_step),
        trajectory.end_time(),
    ):
        root_context.SetTime(t)
        plant.SetPositions(plant_context, trajectory.value(t)[: plant.num_positions()])
        visualizer.ForcedPublish(visualizer_context)

    visualizer.stop_recording()
    return visualizer.get_recording_as_animation()



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestCompassGaitLimitCycle(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    # starts by testing the constraints we already implemented
    # but gives zero points for each test

    @weight(0)
    @timeout_decorator.timeout(3.0)
    def test_time_intervals(self):
        """Time intervals are within bounds"""
        # retrieve variables from notebook
        T = self.notebook_locals["T"]
        h_opt = self.notebook_locals["h_opt"]
        h_min = self.notebook_locals["h_min"]
        h_max = self.notebook_locals["h_max"]

        # check one per time
        tol = 1e-4
        for t in range(T):
            self.assertTrue(
                h_opt[t] >= h_min - tol,
                msg=f"Time interval lower bound violated at time " + f"step {t}.",
            )
            self.assertTrue(
                h_opt[t] <= h_max + tol,
                msg=f"Time interval upper bound violated at time " + f"step {t}.",
            )

    @weight(0)
    @timeout_decorator.timeout(3.0)
    def test_dynamic_feasibility(self):
        """Implicit Euler equations are verified"""
        # retrieve variables from notebook
        T = self.notebook_locals["T"]
        h = self.notebook_locals["h_opt"]
        q = self.notebook_locals["q_opt"]
        qd = self.notebook_locals["qd_opt"]
        qdd = self.notebook_locals["qdd_opt"]
        f = self.notebook_locals["f_opt"]
        manipulator_equations = self.notebook_locals["manipulator_equations"]

        # link configurations, velocities, accelerations using implicit Euler
        for t in range(T):
            np.testing.assert_allclose(
                q[t + 1],
                q[t] + h[t] * qd[t + 1],
                atol=1e-5,
                rtol=1e-5,
                err_msg="Configurations q and velocities qd do not verify "
                + f"the Implicit Euler rule at time step {t}.",
            )
            np.testing.assert_allclose(
                qd[t + 1],
                qd[t] + h[t] * qdd[t],
                atol=1e-5,
                rtol=1e-5,
                err_msg="Velocities qd and accelerations qdd do not verify "
                + f"the Implicit Euler rule at time step {t}.",
            )
            vars = np.concatenate((q[t + 1], qd[t + 1], qdd[t], f[t]))
            manipulator_equations_violation = manipulator_equations(vars)
            np.testing.assert_allclose(
                manipulator_equations_violation,
                np.zeros(q.shape[1]),
                atol=1e-5,
                rtol=1e-5,
                err_msg="Manipulator equations are not verified at time "
                + f"step {t}.",
            )

    @weight(0)
    @timeout_decorator.timeout(3.0)
    def test_impulsive_collision(self):
        """Impulsive collision is verified at the heel strike"""
        # retrieve variables from notebook
        q = self.notebook_locals["q_opt"]
        qd = self.notebook_locals["qd_opt"]
        qd_post = self.notebook_locals["qd_post_opt"]
        imp = self.notebook_locals["imp_opt"]
        reset_velocity_heelstrike = self.notebook_locals["reset_velocity_heelstrike"]

        # link pre and post impact velocities
        vars = np.concatenate((q[-1], qd[-1], qd_post, imp))
        reset_velocity_heelstrike_violation = reset_velocity_heelstrike(vars)
        np.testing.assert_allclose(
            reset_velocity_heelstrike_violation,
            np.zeros(q.shape[1] + imp.size),
            atol=1e-5,
            rtol=1e-5,
            err_msg="Impulsive collision equations are not verified.",
        )

    @weight(0)
    @timeout_decorator.timeout(3.0)
    def test_periodicity(self):
        """Periodicity of the walking gait"""
        # retrieve variables from notebook
        q = self.notebook_locals["q_opt"]
        qd = self.notebook_locals["qd_opt"]
        qd_post = self.notebook_locals["qd_post_opt"]

        # periodicity of the configuration
        np.testing.assert_allclose(
            q[0],
            -q[-1],
            atol=1e-5,
            rtol=1e-5,
            err_msg="Configuration vector q does not verify periodicity "
            + "conditions.",
        )

        # periodicity of the  velocities
        np.testing.assert_allclose(
            qd[0],
            np.array([0, 0, qd_post[2] + qd_post[3], -qd_post[3]]),
            atol=1e-5,
            rtol=1e-5,
            err_msg="Velocity vector qd does not verify periodicity " + "conditions.",
        )

    # tests constraints implemented by the students
    # does not use the same expressions that must be used in the constraint
    # e.g. checks that a linear constraint is verified by evaluating a nonlinear
    # function

    @weight(3)
    @timeout_decorator.timeout(3.0)
    def test_stance_foot_on_ground(self):
        """Stance foot on the ground for all times"""
        # retrieve variables from notebook
        T = self.notebook_locals["T"]
        q = self.notebook_locals["q_opt"]
        for t in range(T + 1):
            self.assertAlmostEqual(
                np.linalg.norm(q[t, :2]),
                0,
                msg=f"Stance foot is not on the ground at time step {t}.",
            )

    @weight(2)
    @timeout_decorator.timeout(3.0)
    def test_swing_foot_on_ground(self):
        """Swing foot on the ground at time zero"""
        # retrieve variables from notebook
        q = self.notebook_locals["q_opt"]
        swing_foot_height = self.notebook_locals["swing_foot_height"]
        self.assertAlmostEqual(
            swing_foot_height(q[0])[0],
            0,
            msg="Swing foot is not on the ground at time zero.",
        )

    @weight(3)
    @timeout_decorator.timeout(3.0)
    def test_no_penetration(self):
        """No penetration of the swing foot in the ground for all times"""
        # retrieve variables from notebook
        T = self.notebook_locals["T"]
        q = self.notebook_locals["q_opt"]
        swing_foot_height = self.notebook_locals["swing_foot_height"]
        tol = 1e-4
        for t in range(T + 1):
            self.assertGreater(
                swing_foot_height(q[t])[0],
                -tol,
                msg=f"Swing foot in penetration with ground at time step {t}.",
            )

    @weight(3)
    @timeout_decorator.timeout(3.0)
    def test_stance_foot_friction(self):
        """Stance-foot contact force in friction cone for all times"""
        # retrieve variables from notebook
        T = self.notebook_locals["T"]
        friction = self.notebook_locals["friction"]
        f = self.notebook_locals["f_opt"]
        tol = 1e-4
        for t in range(T):
            self.assertGreater(
                f[t, 1],
                -tol,
                msg="Stance-foot normal contact force is negative at time "
                + f"step {t}.",
            )
            self.assertLess(
                np.abs(f[t, 0]),
                friction * f[t, 1] + tol,
                msg="Stance-foot contact force outside friction cone at time "
                + f"step {t}.",
            )

    @weight(2)
    @timeout_decorator.timeout(3.0)
    def test_swing_foot_friction(self):
        """Swing-foot impulse in friction cone"""
        # retrieve variables from notebook
        friction = self.notebook_locals["friction"]
        imp = self.notebook_locals["imp_opt"]
        tol = 1e-4
        self.assertGreater(
            imp[1],
            -tol,
            msg="Normal heel-strike impulse is negative at time step {t}.",
        )
        self.assertLess(
            np.abs(imp[0]),
            friction * imp[1] + tol,
            msg="Heel-strike impulse outside friction cone",
        )



import matplotlib.animation as ani
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm

# pydrake imports


def create_animation(J_grid, pi_grid, mesh):
    # initialize figure for animation plus misc settings
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    plt.tight_layout()
    ax[0].axis("off")
    ax[1].axis("off")

    # cost-to-go plot and policy plots
    # overwrites the subplots axes
    ax[0] = fig.add_subplot(121, projection="3d")
    ax[1] = fig.add_subplot(122, projection="3d")
    ax[0].set_title(r"Optimal cost to go   $J^*(\mathbf{x})$")
    ax[1].set_title(r"Optimal policy   $\pi^*(\mathbf{x})$")
    for axi in ax:
        axi.set_xlabel(r"$q$")
        axi.set_ylabel(r"$\dot q$")

    # helper function for the surface plot
    Q, Qdot = np.meshgrid(mesh["q_grid"], mesh["qdot_grid"])

    def plot_surf(ax, Z):
        return ax.plot_surface(Q, Qdot, Z.T, rstride=1, cstride=1, cmap=cm.jet)

    # first frame of the animation
    J_surf = [plot_surf(ax[0], J_grid[:, :, 0])]
    pi_surf = [plot_surf(ax[1], pi_grid[:, :, 0])]

    # video parameters
    frames = 40  # total number of frames
    duration = 10  # seconds
    interval = duration / (frames - 1) * 1000  # milliseconds

    # initialize title to be modified in the callback
    title = fig.text(
        0.5, 0.95, "", fontsize="x-large", bbox={"facecolor": "w"}, ha="center"
    )

    # callback function for the animation
    def update_surf(frame, J_grid, J_surf, pi_grid, pi_surf):
        # iteration to plot, - 1 for base zero
        iters = J_grid.shape[2]
        i = int(frame * (iters - 1) / (frames - 1))

        # update cost-to-go and policy
        J_surf[0].remove()
        pi_surf[0].remove()
        J_surf[0] = plot_surf(ax[0], J_grid[:, :, i])
        pi_surf[0] = plot_surf(ax[1], pi_grid[:, :, i])

        # update title with current iteration
        # use base 1 as above
        title.set_text(f"Value iteration {i+1}")

    # create animation
    animate = ani.FuncAnimation(
        fig,
        update_surf,
        frames=frames,
        interval=interval,
        fargs=(J_grid, J_surf, pi_grid, pi_surf),
    )

    # play video
    plt.close()  # close any open figure
    return animate


# function that plots the trajectory of the
# double integrator in state space (q_sim vs qdot_sim)
def plot_state_trajectory(q_sim, qdot_sim):
    # draw a white dot for the initial conditions
    plt.scatter(
        q_sim[0],
        qdot_sim[0],
        s=100,
        edgecolor="k",
        c="w",
        zorder=3,
        label=r"$\mathbf{x}(0)$",
    )

    # black line for the trajectory in time
    plt.plot(q_sim, qdot_sim, color="k", linewidth=2, label=r"$\mathbf{x}(t)$")

    # misc
    plt.xlabel(r"$q$")
    plt.ylabel(r"$\dot q$")
    plt.legend()


# helper function for plot_policy that evaluates
# the controller output at a given state
def evaluate_controller(controller, q, qdot):
    # get context and set system output (= system state)
    context = controller.CreateDefaultContext()
    controller.get_input_port(0).FixValue(context, (q, qdot))

    # compute input for the double integrator
    u = controller.get_output_port(0).Eval(context)[0]

    return u


# function that produces a level plot
# the policy generated by the passed controller
# it used the grid defined by q_grid and qdot_grid
def plot_policy(q_grid, qdot_grid, controller):
    # evaluate the policy on the grid
    Pi_grid = np.array(
        [
            [evaluate_controller(controller, q, qdot) for qdot in qdot_grid]
            for q in q_grid
        ]
    )

    # level plot the level function
    # note the transpose to align the policy to the grid
    plt.contourf(*np.meshgrid(q_grid, qdot_grid), Pi_grid.T, cmap=cm.jet)

    # add a bar with the color scale on the right
    plt.colorbar(label=r"$\pi^*(\mathbf{x})$")


# function that plots the control signal
# u_sim as a function of the time vector t
def plot_input(u_sim, t_sim, u_lim):
    # plot the bounds for the control signal
    plt.plot(
        t_sim,
        [u_lim[0]] * len(t_sim),
        c="r",
        linestyle="--",
        label="Input limits",
    )
    plt.plot(t_sim, [u_lim[1]] * len(t_sim), c="r", linestyle="--")

    # plot the control signal
    plt.plot(t_sim, u_sim, label="Input from simulation")

    # misc
    plt.xlabel(r"$t$")
    plt.ylabel(r"$u$")
    plt.xlim(min(t_sim), max(t_sim))
    plt.grid(True)
    plt.legend(loc=1)


# overall plot function for the state trajectory,
# controller policy, and input signal
def simulate_and_plot(
    q0, qdot0, sim_time, controller, u_lim, nq=201, nqdot=201, simulate=None
):
    # get trajectories
    q_sim, qdot_sim, u_sim, t_sim = simulate(q0, qdot0, sim_time, controller)

    # state figure
    plt.figure()
    plot_state_trajectory(q_sim, qdot_sim)

    # plot policy only in a square region that
    # contains tightly the trajectory

    # helper function that computes upper and
    # lower bounds, with a small frame, for the
    # passed signal s
    def frame_signal(s, frame=0.1):
        ds = (max(s) - min(s)) * frame
        return [min(s) - ds, max(s) + ds]

    # regrid state space for policy plot
    # this grid must be much finer than the
    # one used for value iteration
    q_grid = np.linspace(*frame_signal(q_sim), nq)
    qdot_grid = np.linspace(*frame_signal(qdot_sim), nqdot)

    # plot the policy from the passed controller
    plot_policy(q_grid, qdot_grid, controller)

    # plot input as a function of time
    plt.figure()
    plot_input(u_sim, t_sim, u_lim)



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import System


class TestDrakeSystems(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_input_and_state(self):
        # note: all prints here go to the output item in the json file

        system = self.notebook_locals["pendulum_system"]

        assert isinstance(system, System), "pendulum_system is not a System"
        assert system.num_input_ports() == 1, "should have 1 input port"
        assert system.get_input_port(0).size() == 1, "the input port is the wrong size"
        assert system.num_continuous_states() == 2, "wrong size for the state vector"
        assert (
            system.num_discrete_state_groups() == 0
        ), "your system should not have discrete states"

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_dynamics(self):
        # note: all prints here go to the output item in the json file

        system = self.notebook_locals["pendulum_system"]
        context = system.CreateDefaultContext()

        q = np.linspace(0, 1, 5)
        v = np.linspace(0, 1, 5)
        u = np.linspace(0, 1, 5)
        qs, vs, us = np.meshgrid(q, v, u)
        for qi, vi, ui in zip(qs.flat, vs.flat, us.flat):
            context.SetContinuousState([qi, vi])
            system.get_input_port(0).FixValue(context, [ui])
            xdot = [vi, ui - vi - 10 * np.sin(qi)]
            np.testing.assert_almost_equal(
                xdot, system.EvalTimeDerivatives(context).CopyToVector()
            )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.solvers import SnoptSolver


class TestFootstepPlanning(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        """
        Since the planner raises an error if the problem is infeasible, it is
        better to solve the optimizations inside each method rather than here in
        the __init__ method. This way if the plan is infeasible, students get
        zero points instead of an error from the test file. Note also that each
        time a test (method of this class) fails a new class is instantiated,
        hence we must re-solve the plans for each test.
        """
        super().__init__(test_name)

        # get locals
        self.notebook_locals = notebook_locals
        Terrain = self.notebook_locals["Terrain"]
        self.footstep_planner = self.notebook_locals["GCS_footstep_planner"]
        self.solver = SnoptSolver()

        # empty solution of the benchmark problems
        self.terrains = {
            "A": Terrain([1, 1, 1, 1]),
            "B": Terrain([1, 1, 1, 0]),
        }
        self.n_steps = 8
        self.step_span = 0.8
        self.plans = {"A": {}, "B": {}}

    def _solve_benchmarks(self):
        """Solves the benchmark problems."""
        for plan in ["A", "B"]:
            position_left, position_right, cost = self.footstep_planner(
                self.terrains[plan], self.step_span, True, self.solver
            )
            self.plans[plan]["vars"] = [position_left, position_right]
            self.plans[plan]["cost"] = (
                np.linalg.norm(
                    position_left[1:]
                    - position_left[:-1]
                    + position_right[1:]
                    - position_right[:-1]
                )
                ** 2
            )

    @weight(8)
    @timeout_decorator.timeout(30.0)
    def test_relative_position_limits(self):
        """Tests the constraints on the relative position of the feet"""
        self._solve_benchmarks()

        # center to corner vector for the square with side step_span
        # (plus tolerance for the checks)
        tol = 1e-2
        c2c = np.ones(2) * self.step_span / 2 + tol

        # check constraint for both terrains
        for plan in ["A", "B"]:
            position_left, position_right = self.plans[plan]["vars"][:2]
            relative_position = position_left - position_right
            for t in range(self.n_steps):
                msg = (
                    f"Error at foot step {t} when solving the problem "
                    + f"for terrain_{plan}: one foot is not contained in a "
                    + f"square of side {self.step_span} centered at the "
                    + "other foot."
                )
                np.testing.assert_array_less(relative_position[t], c2c, msg)
                np.testing.assert_array_less(-relative_position[t], c2c, msg)

    @weight(8)
    @timeout_decorator.timeout(30.0)
    def test_minimize_step_length(self):
        """
        Tests that the positioning of the feet minimized the sum of the
        square of the two norm of the step lengths.
        """
        self._solve_benchmarks()

        # check constraint for both terrains
        target_cost = {"A": 5.3199999999, "B": 6.4799999999}
        for plan in ["A", "B"]:
            # same convergence check as in
            # https://github.com/RobotLocomotion/drake/blob/88c93118df507777eb3f99628d1aa7c808f81f49/solvers/branch_and_bound.cc#L711
            # with double the tolerances from
            # https://github.com/RobotLocomotion/drake/blob/4ee674e7931527df838bd33e79cf2f4dad57bd20/solvers/branch_and_bound.h#L674
            atol = 2e-2
            rtol = 2e-2
            gap = self.plans[plan]["cost"] - target_cost[plan]
            atol_convergence = gap <= atol
            rtol_convergence = gap / target_cost[plan] <= rtol
            self.assertTrue(
                atol_convergence or rtol_convergence,
                f"Target cost of {target_cost[plan]} is not achieved for "
                + f"terrain_{plan}, with absolute tolerance {atol} and "
                + f"relative tolerance {rtol}.",
            )



from .jupyter import running_as_notebook
from .multibody import ManipulatorDynamics
from .plot_utils import plot_2d_phase_portrait
from .pyplot_utils import SliderSystem

# Note: can't import running_as_test here.  Will get a copy instead of a
# reference.
from .utils import ConfigureParser, FindResource



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestOrbitalTransfer(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    def trajopt_feasibility(self):
        """Test feasibility of the trajopt solution"""
        # load locals
        universe = self.notebook_locals["universe"]
        thrust = self.notebook_locals["thrust_opt"]
        state = self.notebook_locals["state_opt"]
        time_steps = self.notebook_locals["time_steps"]
        time_interval = self.notebook_locals["time_interval"]

        # initial state in Earth orbit
        residuals = universe.constraint_state_to_orbit(state[0], "Earth")
        np.testing.assert_array_almost_equal(
            residuals,
            np.zeros(residuals.shape),
            err_msg="The initial state in not on the Earth orbit.",
        )

        # final state in Mars orbit
        residuals = universe.constraint_state_to_orbit(state[-1], "Mars")
        np.testing.assert_array_almost_equal(
            residuals,
            np.zeros(residuals.shape),
            # decimal=4,
            err_msg="The final state in not on the Mars orbit.",
        )

        # discretized dynamics
        for t in range(time_steps):
            residuals = universe.rocket_discrete_dynamics(
                state[t], state[t + 1], thrust[t], time_interval
            )
            np.testing.assert_array_almost_equal(
                residuals,
                np.zeros(residuals.shape),
                err_msg=f"Discrete dynamics at time step {t} is not verified.",
            )

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test_thrust_limits(self):
        """Test thrust limits"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals["universe"]
        thrust = self.notebook_locals["thrust_opt"]
        time_steps = self.notebook_locals["time_steps"]

        # check thrust limits
        tol = 1e-3
        thrust_limit = universe.rocket.thrust_limit
        for t in range(time_steps):
            self.assertTrue(
                np.linalg.norm(thrust[t]) <= thrust_limit + tol,
                msg=f"Thrust limit violated at time step {t}.",
            )

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test_velocity_limits(self):
        """Test velocity limits"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals["universe"]
        velocity = self.notebook_locals["state_opt"][:, 2:]
        time_steps = self.notebook_locals["time_steps"]

        # check velocity limits
        tol = 1e-3
        velocity_limit = universe.rocket.velocity_limit
        for t in range(time_steps + 1):
            self.assertTrue(
                np.linalg.norm(velocity[t]) <= velocity_limit + tol,
                msg=f"Velocity limit violated at time step {t}.",
            )

    @weight(6)
    @timeout_decorator.timeout(1.0)
    def test_asteroid_collision_limits(self):
        """Test asteroid collision"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        universe = self.notebook_locals["universe"]
        state = self.notebook_locals["state_opt"]
        time_steps = self.notebook_locals["time_steps"]
        n_asteroids = self.notebook_locals["n_asteroids"]

        # avoid collision with asteroids
        tol = 1e-3
        for i in range(n_asteroids):
            asteroid_orbit = universe.get_planet(f"Asteroid_{i}").orbit
            for t in range(time_steps + 1):
                p = universe.position_wrt_planet(state[t], f"Asteroid_{i}")
                self.assertTrue(
                    p.dot(p) >= asteroid_orbit**2 - tol,
                    msg=f"Collision with asteroid {i} at time step {t}.",
                )

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test_fuel_consumption_limits(self):
        """Test fuel consumption"""
        # check feasibility first
        self.trajopt_feasibility()

        # load locals
        self.notebook_locals["universe"]
        thrust = self.notebook_locals["thrust_opt"]
        time_interval = self.notebook_locals["time_interval"]

        # check maximum consumption
        consumption = time_interval * sum(t.dot(t) for t in thrust)
        self.assertTrue(
            consumption <= 250,
            msg=f"Fuel consumption is {consumption}, greater than 250.",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import Solve


class TestVanDerPol(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_linearization(self):
        """Test linearization matrix A"""
        A = self.notebook_locals["A"]
        target_A = np.array([[0, -1], [1, -1]])
        assert A.shape == (
            2,
            2,
        ), f"A must be a 2x2 matrix. Given A is {A.shape}"  # yapf: disable
        err = np.linalg.norm(A - target_A)
        self.assertLessEqual(err, 1e-8, "incorrect linearization matrix A")

    @weight(1)
    @timeout_decorator.timeout(5.0)
    def test_lyapunov_matrices(self):
        """Test lyapunov matrices"""
        A = self.notebook_locals["A"]
        Q = self.notebook_locals["Q"]
        P = self.notebook_locals["P"]
        assert Q.shape == (
            2,
            2,
        ), f"Q must be a 2x2 matrix. Given Q is {Q.shape}"  # yapf: disable
        assert P.shape == (
            2,
            2,
        ), f"P must be a 2x2 matrix. Given P is {P.shape}"  # yapf: disable
        target_Q = -(A.T @ P + P @ A)
        np.testing.assert_array_almost_equal(
            Q,
            target_Q,
            err_msg="A'P + PA is not equal to -Q.",
        )

    @weight(3)
    @timeout_decorator.timeout(5.0)
    def test_method_1(self):
        """Test level set for Method 1"""
        # test level set value up to first digit
        rho_method_1 = self.notebook_locals["rho_method_1"]
        self.assertEqual(
            hash(self.round(rho_method_1, 1)),
            691752902764107778,
            "The level set from Method 1 is incorrect.",
        )

        # test that the line search converged
        is_verified = self.notebook_locals["is_verified"]
        self.assertTrue(
            is_verified(rho_method_1),
            "Could not verify the level set rho_method_1 with the function"
            + "is_verified(rho).",
        )
        rho_step = self.notebook_locals["rho_step"]
        self.assertAlmostEqual(
            rho_step,
            0.01,
            msg="rho_step is different from the predefined value of 0.01.",
        )
        self.assertFalse(
            is_verified(rho_method_1 + rho_step),
            "The function is_verified(rho) could verify the level set"
            + "rho_method_1 + rho_step: line search is not complete.",
        )

    @weight(5)
    @timeout_decorator.timeout(5.0)
    def test_method_2(self):
        """Test level set for Method 2"""
        # test level set value up to fourth digit
        # (note that the degree of the polynomial does not affect precision
        # in this particular problem)
        rho_method_2 = self.notebook_locals["rho_method_2"]
        self.assertEqual(
            hash(self.round(rho_method_2, 4)),
            702129196305569794,
            "The level set from Method 2 is incorrect.",
        )

        # try to solve again the optimization problem
        # and test that the optimal value of rho is the same
        # to retrieve rho, find the only optimization variable in the cost
        prog = self.notebook_locals["prog2"]
        costs = prog.GetAllCosts()
        self.assertEqual(len(costs), 1, "prog2 has more than one cost function.")
        variables = prog.GetAllCosts()[0].variables()
        self.assertEqual(len(variables), 1, "The cost function of prog2 is incorrect.")
        rho = variables[0]
        result = Solve(prog)
        self.assertAlmostEqual(
            result.GetSolution(rho),
            rho_method_2,
            msg="Solving prog2 we got a different value of rho_method_2.",
        )

    @weight(5)
    @timeout_decorator.timeout(5.0)
    def test_method_3(self):
        """Test level set for Method 3"""
        # test level set value up to fourth digit
        # (note that the degree of the polynomial does not affect precision
        # in this particular problem)
        rho_method_3 = self.notebook_locals["rho_method_3"]
        self.assertEqual(
            hash(self.round(rho_method_3, 4) * 2),
            1404258392611139588,
            "The level set from Method 3 is incorrect.",
        )

        # try to solve again the optimization problem
        # and test that the optimal value of rho is the same
        # to retrieve rho, find the only optimization variable in the cost
        prog = self.notebook_locals["prog3"]
        costs = prog.GetAllCosts()
        self.assertEqual(len(costs), 1, "prog3 has more than one cost function.")
        variables = prog.GetAllCosts()[0].variables()
        self.assertEqual(len(variables), 1, "The cost function of prog3 is incorrect.")
        rho = variables[0]
        result = Solve(prog)
        self.assertAlmostEqual(
            result.GetSolution(rho),
            rho_method_3,
            msg="Solving prog3 we got a different value of rho_method_3.",
        )

    @staticmethod
    def round(n, decimals=0):
        multiplier = 10**decimals
        return round(n * multiplier) / multiplier



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestOneDHopper(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(1.0)
    def test_mechanical_energy(self):
        """Test mechanical energy"""
        # retrieve energy function
        mechanical_energy = self.notebook_locals["mechanical_energy"]

        # evaluate energy at N random points
        N = 100
        np.random.seed(0)
        state = np.random.randn(4, N)
        delta = np.random.randn(N)
        E = mechanical_energy(state, delta)

        # check size of E
        self.assertFalse(
            isinstance(E, (float, int)),
            msg=f"The function mechanical_energy does not take vectorized " + "inputs.",
        )
        self.assertTrue(
            E.size == N,
            msg=f"Passing an array of shape (4, {N}) for the states and an "
            + f"array of shape ({N},) for delta, got back an array of size "
            + f"{E.size} from the function mechanical_energy. The output array "
            + f"must have size {N} instead.",
        )

        # compare with target values
        E_target = np.array(
            [
                1035.88403634,
                3066.05294704,
                226.59829248,
                1132.19808917,
                4519.60941109,
                2571.77722312,
                179.75301624,
                218.267293,
                1518.71951559,
                285.28744414,
                126.93879604,
                1623.09759602,
                761.6724944,
                1439.16285815,
                1123.31075613,
                1150.29073766,
                1039.18526308,
                -22.85241161,
                2344.88095522,
                3406.4444863,
                586.89991222,
                194.75259958,
                1710.76292998,
                985.57408993,
                593.67007518,
                -156.80729216,
                129.76883127,
                397.72717689,
                179.93169091,
                392.49996925,
                784.32597859,
                433.87021831,
                -81.97555113,
                -164.62661332,
                1261.32659852,
                518.81015917,
                1127.06741275,
                171.04599127,
                678.92178616,
                476.58699247,
                1653.76175357,
                111.44797797,
                -178.80377977,
                562.21356062,
                5279.60802163,
                321.88033307,
                -33.49681192,
                113.06257362,
                2018.19847436,
                147.46771374,
                534.50643625,
                3271.78401092,
                -21.5353477,
                15.30134453,
                27.29342195,
                963.63275164,
                240.49348165,
                5417.05008921,
                384.4609627,
                78.46055822,
                445.4086912,
                433.45750051,
                53.56942879,
                -149.21833514,
                1521.20337472,
                5268.5937194,
                1737.91194436,
                2767.73189444,
                520.04810383,
                170.84779081,
                688.84932501,
                86.93212697,
                323.27284373,
                786.95926209,
                283.83230408,
                1338.02751306,
                33.00346265,
                1167.80440843,
                211.53932223,
                163.70938173,
                -40.39089787,
                149.60356967,
                1255.61900843,
                3329.55930887,
                1274.00886131,
                9255.17733426,
                207.51077886,
                171.65671047,
                190.63432638,
                1693.4769287,
                2643.46635344,
                139.37393331,
                613.39405767,
                111.57079258,
                5226.36517526,
                111.22451849,
                15.62008105,
                251.80124408,
                692.01834266,
                850.05120716,
            ]
        )
        np.testing.assert_array_almost_equal(
            E_target,
            E.flatten(),
            err_msg="The function mechanical_energy does not return the "
            + "correct values when evaluated at random points.",
        )

    @weight(10)
    @timeout_decorator.timeout(15.0)
    def test_preload_controller(self):
        """Test preload controller"""
        # retrieve controller and simulator
        PreloadController = self.notebook_locals["PreloadController"]
        simulate = self.notebook_locals["simulate"]

        # number of random simulations
        N = 10

        # duration of each simulation, just enough to ensure
        # convergence independently of the initial condition
        duration = 6

        # final window of time in which convergence must be achieved
        tail = 2

        np.random.seed(1)
        for i in range(N):
            # desired hopping height from .5 to 1.5
            h = np.random.rand() + 0.5

            # initial body height from 0 to 1
            zb0 = np.random.rand()
            initial_state = np.array([zb0, 0, 0, 0])

            # instantiate controller
            preload_controller = PreloadController(h)

            # run simulation
            logger = simulate(initial_state, duration, preload_controller)[1]

            # unpack logger
            time = logger.sample_times()
            zb = logger.data()[0]

            # get body height for the final window of time
            zb_tail = zb[np.where(time >= duration - tail)[0]]

            # ensure that in final 2 seconds of simulation the body
            # is always below h plus a tolerance
            tol = 0.1
            self.assertTrue(
                max(zb_tail) < h + tol,
                msg=f"Starting from the state {initial_state} and having a "
                + f"target hopping height of {h}, the hopper body still hops "
                + f"above {h} after {duration - tail} seconds.",
            )

            # ensure that in final 2 seconds of simulation the body achieves
            # the hopping height h at least once
            self.assertTrue(
                max(zb_tail) > h - tol,
                msg=f"Starting from the state {initial_state} and having a "
                + f"target hopping height of {h}, the hopper body does not "
                + f"reach the height {h} after {duration - tail} seconds.",
            )



import numpy as np
import numpy.typing as npt
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import JointActuatorIndex, MultibodyForces_
from pydrake.systems.framework import Context


def ManipulatorDynamics(
    plant: MultibodyPlant,
    q: npt.NDArray,
    v: npt.NDArray = None,
    context: Context = None,
):
    """Returns the components of the manipulator equations -- M(q), C(q,v), tauG(q), B(q), and tauExt(q, v) -- for a given MultibodyPlant.

    Args:
        plant: The MultibodyPlant for which to compute the manipulator equations.
        q: The generalized positions.
        v: The generalized velocities.  If None, the velocities are taken from the
           context.
        context: The Context to use for the computation.  If None, a new default
                 Context will be created.
    """
    if context is None:
        context = plant.CreateDefaultContext()
    plant.SetPositions(context, q)
    if v is not None:
        plant.SetVelocities(context, v)
    M = plant.CalcMassMatrixViaInverseDynamics(context)
    Cv = plant.CalcBiasTerm(context)
    tauG = plant.CalcGravityGeneralizedForces(context)
    B = plant.MakeActuationMatrix()
    forces = MultibodyForces_(plant)
    plant.CalcForceElementsContribution(context, forces)
    # TODO(russt): add in contact forces to tauExt.
    tauExt = plant.CalcGeneralizedForces(context, forces) - tauG

    return (M, Cv, tauG, B, tauExt)


def MakePidStateProjectionMatrix(plant: MultibodyPlant):
    """Given a MultibodyPlant, returns a selection matrix, S, such that [q_a; v_a] = S @ [q; v], where q_a and v_a are the actuated positions and velocities and q, v, are the full state. This can be passed to e.g. the `state_projection` argument of the PidController constructor.

    Args:
        plant: The MultibodyPlant for the projection matrix.
    """
    num_q = plant.num_positions()
    num_v = plant.num_velocities()
    num_u = plant.num_actuators()
    S = np.zeros((2 * num_u, num_q + num_v))
    j = 0
    for i in range(plant.num_actuators()):
        actuator = plant.get_joint_actuator(JointActuatorIndex(i))
        assert actuator.num_inputs() == 1
        joint = actuator.joint()
        S[j, joint.position_start()] = 1
        S[num_u + j, num_q + joint.velocity_start()] = 1
        j = j + 1
    return S



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestCartPolesURDF(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test_x_star_single_state(self):
        """Test x_star for single pendulum"""
        # note: all prints here go to the output item in the json file
        x_star_obtained = self.notebook_locals["x_star_single_pendulum"]
        assert (
            len(x_star_obtained) == 4
        ), "x_star for single pendulum must have dimension 4"  # yapf: disable
        x_star_target = np.array([0, np.pi, 0, 0])
        np.testing.assert_array_almost_equal(
            x_star_obtained,
            x_star_target,
            err_msg="unstable equilibrium for single pendulum does not match",
        )

    @weight(1)
    @timeout_decorator.timeout(1.0)
    def test_x_star_double_state(self):
        """Test x_star for double pendulum"""
        # note: all prints here go to the output item in the json file
        x_star_obtained = self.notebook_locals["x_star_double_pendulum"]
        assert (
            len(x_star_obtained) == 6
        ), "x_star for double pendulum must have dimension 6"  # yapf: disable
        x_star_target = np.array([0, np.pi, np.pi, 0, 0, 0])
        np.testing.assert_array_almost_equal(
            x_star_obtained,
            x_star_target,
            err_msg="unstable equilibrium for double pendulum does not match",
        )

    @weight(3)
    @timeout_decorator.timeout(10.0)
    def test_single_pendulum_urdf(self):
        """Test single-pendulum cart-pole dynamics"""
        # note: all prints here go to the output item in the json file
        obtained_single_pend_urdf = self.notebook_locals["single_pendulum_urdf"]
        cartpole_fn = self.notebook_locals["cartpole_balancing"]
        x_star = np.array([0, np.pi, 0, 0])
        Q = np.diag((10.0, 10.0, 1.0, 1.0))
        R = np.eye(1)
        state_lst = np.array(
            [
                [0.05, np.pi * 0.95, -0.01, 0.01 * np.pi],
                [-0.01, np.pi * 1.1, 0.1, 0.0],
                [0.2, np.pi * 0.5, 0.01, 0.0],
                [0.0, np.pi, 0.0, 0.0],
                [-2, np.pi * 1.2, 0.0, -0.1 * np.pi],
            ]
        )
        diagram = cartpole_fn(obtained_single_pend_urdf, x_star, Q, R)
        cartpole = diagram.GetSubsystemByName("cartpole")
        context = cartpole.CreateDefaultContext()
        cartpole.get_actuation_input_port().FixValue(context, [0])
        result_lst = []
        for state in state_lst:
            context.SetContinuousState(state)
            result_lst.append(cartpole.EvalTimeDerivatives(context).CopyToVector())
        result = np.array(result_lst)
        target_derivatives = np.array(
            [
                [
                    -1.00000000e-02,
                    3.14159265e-02,
                    -1.47937117e00,
                    -2.99577976e00,
                ],
                [1.00000000e-01, 0.00000000e00, 2.63177456e00, 5.53442305e00],
                [
                    1.00000000e-02,
                    0.00000000e00,
                    3.00344627e-16,
                    -9.81000000e00,
                ],
                [
                    0.00000000e00,
                    0.00000000e00,
                    -1.20137851e-15,
                    -2.40275702e-15,
                ],
                [0.00000000e00, -3.14159265e-01, 3.42396821e00, 8.53622180e00],
            ]
        )
        np.testing.assert_array_almost_equal(
            result,
            target_derivatives,
            decimal=6,
            err_msg="the dynamics for the single pendulum do not match",
        )

    @weight(3)
    @timeout_decorator.timeout(10.0)
    def test_double_pendulum_urdf(self):
        """Test double-pendulum cart-pole dynamics"""
        # note: all prints here go to the output item in the json file
        obtained_double_pend_urdf = self.notebook_locals["double_pendulum_urdf"]
        cartpole_fn = self.notebook_locals["cartpole_balancing"]
        x_star = np.array([0, np.pi, np.pi, 0, 0, 0])
        Q = np.diag((10.0, 10.0, 10.0, 1.0, 1.0, 1.0))
        R = np.eye(1)
        state_lst = np.array(
            [
                [0.05, np.pi * 0.95, np.pi / 2, -0.01, 0.01 * np.pi, 0.0],
                [-0.01, np.pi * 1.1, -np.pi * 0.1, 0.1, 0.0, -0.2 * np.pi],
                [0.2, np.pi * 0.5, np.pi / 8, 0.01, 0.0, 0.0],
                [0.0, np.pi, np.pi, 0.0, 0.0, 0.0],
                [-2, np.pi * 1.2, 0.01, 0.2, 0.0, -0.1 * np.pi],
            ]
        )
        diagram = cartpole_fn(obtained_double_pend_urdf, x_star, Q, R)
        cartpole = diagram.GetSubsystemByName("cartpole")
        context = cartpole.CreateDefaultContext()
        cartpole.get_actuation_input_port().FixValue(context, [0])
        result_lst = []
        for state in state_lst:
            context.SetContinuousState(state)
            result_lst.append(cartpole.EvalTimeDerivatives(context).CopyToVector())
        result = np.array(result_lst)
        target_derivatives = np.array(
            [
                [
                    -1.00000000e-02,
                    3.14159265e-02,
                    0.00000000e00,
                    -1.47937117e00,
                    -2.99627324e00,
                    -6.46053778e00,
                ],
                [
                    1.00000000e-01,
                    0.00000000e00,
                    -6.28318531e-01,
                    4.57233893e00,
                    9.61528333e00,
                    -5.04294440e00,
                ],
                [
                    1.00000000e-02,
                    0.00000000e00,
                    0.00000000e00,
                    1.61809378e-16,
                    -9.81000000e00,
                    9.81000000e00,
                ],
                [
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    -2.40275702e-15,
                    -2.40275702e-15,
                    0.00000000e00,
                ],
                [
                    2.00000000e-01,
                    0.00000000e00,
                    -3.14159265e-01,
                    5.55139729e00,
                    1.02096347e01,
                    -1.03040792e01,
                ],
            ]
        )
        np.testing.assert_array_almost_equal(
            result,
            target_derivatives,
            decimal=6,
            err_msg="the dynamics for the double pendulum do not match",
        )

    @weight(2)
    @timeout_decorator.timeout(10.0)
    def test_single_pendulum_lqr(self):
        """Test single-pendulum cart-pole LQR states"""
        # note: all prints here go to the output item in the json file
        obtained_single_pend_urdf = self.notebook_locals["single_pendulum_urdf"]
        cartpole_fn = self.notebook_locals["cartpole_balancing"]
        x_star = np.array([0, np.pi, 0, 0])
        Q = np.diag((10.0, 10.0, 1.0, 1.0))
        R = np.eye(1)
        state_lst = np.array(
            [
                [0.05, np.pi * 0.95, -0.01, 0.01 * np.pi],
                [-0.01, np.pi * 1.1, 0.1, 0.0],
                [0.2, np.pi * 0.5, 0.01, 0.0],
                [0.0, np.pi, 0.0, 0.0],
                [-2, np.pi * 1.2, 0.0, -0.1 * np.pi],
            ]
        )
        diagram = cartpole_fn(obtained_single_pend_urdf, x_star, Q, R)
        result_lst = []
        context = diagram.CreateDefaultContext()
        for state in state_lst:
            context.SetContinuousState(state)
            result_lst.append(diagram.EvalTimeDerivatives(context).CopyToVector())
        result = np.array(result_lst)
        target_derivatives = np.array(
            [
                [
                    -1.00000000e-02,
                    3.14159265e-02,
                    7.54162904e00,
                    5.91415697e00,
                ],
                [
                    1.00000000e-01,
                    0.00000000e00,
                    -1.45222333e01,
                    -1.07800079e01,
                ],
                [1.00000000e-02, 0.00000000e00, 4.85651106e01, -9.81000000e00],
                [
                    0.00000000e00,
                    0.00000000e00,
                    -1.20137851e-15,
                    -2.40275702e-15,
                ],
                [
                    0.00000000e00,
                    -3.14159265e-01,
                    -2.61730991e01,
                    -1.54083086e01,
                ],
            ]
        )
        np.testing.assert_array_almost_equal(
            result,
            target_derivatives,
            decimal=4,
            err_msg="the dynamics for the single pendulum LQR do not match",
        )

    @weight(2)
    @timeout_decorator.timeout(10.0)
    def test_double_pendulum_lqr(self):
        """Test double-pendulum cart-pole LQR states"""
        # note: all prints here go to the output item in the json file
        obtained_double_pend_urdf = self.notebook_locals["double_pendulum_urdf"]
        diagram_fn = self.notebook_locals["cartpole_balancing"]
        x_star = np.array([0, np.pi, np.pi, 0, 0, 0])
        Q = np.diag((10.0, 10.0, 10.0, 1.0, 1.0, 1.0))
        R = np.eye(1)
        diagram = diagram_fn(obtained_double_pend_urdf, x_star, Q, R)
        context = diagram.CreateDefaultContext()
        state_lst = np.array(
            [
                [0.05, np.pi * 0.95, np.pi / 2, -0.01, 0.01 * np.pi, 0.0],
                [-0.01, np.pi * 1.1, -np.pi * 0.1, 0.1, 0.0, -0.2 * np.pi],
                [0.2, np.pi * 0.5, np.pi / 8, 0.01, 0.0, 0.0],
                [0.0, np.pi, np.pi, 0.0, 0.0, 0.0],
                [-2, np.pi * 1.2, 0.01, 0.2, 0.0, -0.1 * np.pi],
            ]
        )
        result_lst = []
        context = diagram.CreateDefaultContext()
        for state in state_lst:
            context.SetContinuousState(state)
            result_lst.append(diagram.EvalTimeDerivatives(context).CopyToVector())
        result = np.array(result_lst)
        target_derivatives = np.array(
            [
                [
                    -1.00000000e-02,
                    3.14159265e-02,
                    0.00000000e00,
                    -4.68965989e02,
                    -4.64727356e02,
                    5.28401564e02,
                ],
                [
                    1.00000000e-01,
                    0.00000000e00,
                    -6.28318531e-01,
                    -9.02354495e02,
                    -7.77737857e02,
                    9.40419695e02,
                ],
                [
                    1.00000000e-02,
                    0.00000000e00,
                    0.00000000e00,
                    -3.32561210e02,
                    9.27487605e01,
                    1.29268844e02,
                ],
                [
                    0.00000000e00,
                    0.00000000e00,
                    0.00000000e00,
                    -2.40275702e-15,
                    -2.40275702e-15,
                    0.00000000e00,
                ],
                [
                    2.00000000e-01,
                    0.00000000e00,
                    -3.14159265e-01,
                    -5.40431505e02,
                    -4.34708484e02,
                    4.28196374e02,
                ],
            ]
        )
        np.testing.assert_array_almost_equal(
            result,
            target_derivatives,
            decimal=4,
            err_msg="the dynamics for the double pendulum LQR do not match",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestSOSandPSD(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(2)
    @timeout_decorator.timeout(1.0)
    def test_sos(self):
        """Test SOS functions"""
        # load the response from locals
        obtained = self.notebook_locals["is_function_sos"]
        assert len(obtained) == 3, "is_functions_sos must have length 3"

        # multiplying arbitrary value to mask the difference with the target
        err = np.linalg.norm(obtained - np.array([0.0, 1.0, 0.0])) * 123.45
        self.assertLessEqual(err, 1e-6, "incorrect SOS results")

    @weight(2)
    @timeout_decorator.timeout(1.0)
    def test_psd(self):
        """Test PSD functions"""
        # load the response from locals
        obtained = self.notebook_locals["is_function_psd"]
        assert len(obtained) == 3, "is_functions_psd must have length 3"

        # multiplying arbitrary value to mask the difference with the target
        err = np.linalg.norm(obtained - np.array([1.0, 1.0, 1.0])) * 123.45
        self.assertLessEqual(err, 1e-6, "incorrect PSD results")



import unittest

import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestGliderSysid(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(20.0)
    def test_lq_fit(self):
        """Test least squares fitting"""
        # note: all prints here go to the output item in the json file

        # the best config and best score should have the right values
        best_config = self.notebook_locals["best_config"]
        best_score = self.notebook_locals["best_score"]
        self.assertFalse(best_config != 6, "The chosen basis set is incorrect")
        self.assertFalse(best_score != 63831, "The best score is incorrect")

    @weight(3)
    @timeout_decorator.timeout(20.0)
    def test_basis_dynamics(self):
        """Test basis function dynamics"""
        # note: all prints here go to the output item in the json file
        trajectories_fit = self.notebook_locals["trajectories_fit"]
        last_x = trajectories_fit[0][-1][0]
        last_z = trajectories_fit[0][-1][1]
        last_th = trajectories_fit[0][-1][2]
        self.assertLessEqual(last_x, 3.36, msg="Simulated trajectory x is incorrect")
        self.assertGreaterEqual(last_x, 3.34, msg="Simulated trajectory x is incorrect")
        self.assertLessEqual(last_z, 5.10, msg="Simulated trajectory z is incorrect")
        self.assertGreaterEqual(last_z, 4.99, msg="Simulated trajectory z is incorrect")
        self.assertLessEqual(
            last_th, -0.52, msg="Simulated trajectory theta is incorrect"
        )
        self.assertGreaterEqual(
            last_th, -0.54, msg="Simulated trajectory theta is incorrect"
        )



import matplotlib.pyplot as plt
import numpy as np
from IPython.display import HTML, display
from matplotlib.animation import FuncAnimation
from matplotlib.patches import Rectangle


# helper function that plots a rectangle with given center, width, and height
def plot_rectangle(center, width, height, ax=None, frame=0.1, **kwargs):
    # make black the default edgecolor
    if not "edgecolor" in kwargs:
        kwargs["edgecolor"] = "black"

    # make transparent the default facecolor
    if not "facecolor" in kwargs:
        kwargs["facecolor"] = "none"

    # get current plot axis if one is not given
    if ax is None:
        ax = plt.gca()

    # get corners
    c2c = np.array([width, height]) / 2
    bottom_left = center - c2c
    top_right = center + c2c

    # plot rectangle
    rect = Rectangle(bottom_left, width, height, **kwargs)
    ax.add_patch(rect)

    # scatter fake corners to update plot limits (bad looking but compact)
    ax.scatter(*bottom_left, s=0)
    ax.scatter(*top_right, s=0)

    # make axis scaling equal
    ax.set_aspect("equal")

    return rect


# helper function that generates an animation of planned footstep positions
def animate_footstep_plan(
    terrain, step_span, position_left, position_right, title=None
):
    # initialize figure for animation
    fig, ax = plt.subplots()

    # plot stepping stones
    terrain.plot(title=title, ax=ax)

    # initial position of the feet
    left_foot = ax.scatter(0, 0, color="r", zorder=3, label="Left foot")
    right_foot = ax.scatter(0, 0, color="b", zorder=3, label="Right foot")

    # initial step limits
    left_limits = plot_rectangle(
        [0, 0],  # center
        step_span,  # width
        step_span,  # eight
        ax=ax,
        edgecolor="b",
        label="Left-foot limits",
    )
    right_limits = plot_rectangle(
        [0, 0],  # center
        step_span,  # width
        step_span,  # eight
        ax=ax,
        edgecolor="r",
        label="Right-foot limits",
    )

    # misc settings
    plt.close()
    ax.legend(loc="upper left", bbox_to_anchor=(0, 1.3), ncol=2)

    def animate(n_steps):
        # scatter feet
        left_foot.set_offsets(position_left[n_steps])
        right_foot.set_offsets(position_right[n_steps])

        # limits of reachable set for each foot
        c2c = np.ones(2) * step_span / 2
        right_limits.set_xy(position_left[n_steps] - c2c)
        left_limits.set_xy(position_right[n_steps] - c2c)

    # create ad display animation
    n_steps = position_left.shape[0]
    ani = FuncAnimation(fig, animate, frames=n_steps, interval=1e3)
    display(HTML(ani.to_jshtml()))



import numpy as np
from pydrake.systems.framework import Context, PortDataType
from pydrake.systems.pyplot_visualizer import PyPlotVisualizer


class PendulumVisualizer(PyPlotVisualizer):
    a1 = 0.75
    ac1 = 0.75
    av = np.linspace(0, np.pi, 20)
    rb = 0.03
    hb = 0.07
    aw = 0.01
    base_x = rb * np.concatenate(([1], np.cos(av), [-1]))
    base_y = np.concatenate(([-hb], rb * np.sin(av), [-hb]))
    arm_x = np.concatenate((aw * np.sin(av - np.pi / 2), aw * np.sin(av + np.pi / 2)))
    arm_y = np.concatenate(
        (aw * np.cos(av - np.pi / 2), -a1 + aw * np.cos(av + np.pi / 2))
    )

    def __init__(self, ax=None, show=None):
        PyPlotVisualizer.__init__(self, ax=ax, show=show)
        self.set_name("pendulum_visualizer")
        self.DeclareInputPort("state", PortDataType.kVectorValued, 2)

        self.ax.set_xlim([-1.2, 1.2])
        self.ax.set_ylim([-1.2, 1.2])
        self.base = self.ax.fill(
            self.base_x,
            self.base_y,
            zorder=1,
            facecolor=(0.3, 0.6, 0.4),
            edgecolor="k",
        )

        # arm_x and arm_y are closed (last element == first element), but don't
        # pass the last element to the fill command, because it gets closed
        # anyhow (and we want the sizes to match for the update).
        self.arm = self.ax.fill(
            self.arm_x[0:-1],
            self.arm_y[0:-1],
            zorder=0,
            facecolor=(0.9, 0.1, 0),
            edgecolor="k",
        )
        self.center_of_mass = self.ax.plot(
            0, -self.ac1, zorder=1, color="b", marker="o", markersize=14
        )

    def draw(self, context):
        if isinstance(context, Context):
            theta = self.EvalVectorInput(context, 0).get_value()[0]
            self.ax.set_title("t = {:.1f}".format(context.get_time()))
        else:
            theta = context[0]
            self.ax.set_title("")

        path = self.arm[0].get_path()
        path.vertices[:, 0] = self.arm_x * np.cos(theta) - self.arm_y * np.sin(theta)
        path.vertices[:, 1] = self.arm_x * np.sin(theta) + self.arm_y * np.cos(theta)
        self.center_of_mass[0].set_data(
            self.ac1 * np.sin(theta), -self.ac1 * np.cos(theta)
        )



import unittest

import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestMinimumTime(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(1.0)
    def test_q1_policy_closed_form(self):
        """Test closed form policy"""
        # note: all prints here go to the output item in the json file
        policy_closed_form = self.notebook_locals["policy_closed_form"]
        self.assertEqual(
            policy_closed_form(q=0, qdot=0),
            0,
            "The optimal control at [q=0.0,qd=0.0] should be 0.",
        )

        qdot = -10.0
        q_s = [5.0, -10.0, 0.0]
        for q in q_s:
            self.assertEqual(
                policy_closed_form(q, qdot),
                1,
                "Incorrect control applied at \
                              [q=%2.2f, qd=%2.2f]."
                % (q, qdot),
            )

        qdot = 15.0
        for q in q_s:
            self.assertEqual(
                policy_closed_form(q, qdot),
                -1,
                "Incorrect control applied at \
                             [q=%2.2f, qd=%2.2f]."
                % (q, qdot),
            )
